/*
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */
#ifndef CAOSDB_TRANSACTION_H
#define CAOSDB_TRANSACTION_H
#include "boost/log/core/record.hpp"                  // for record
#include "boost/log/sources/record_ostream.hpp"       // for basic_record_o...
#include "boost/preprocessor/seq/limits/enum_256.hpp" // for BOOST_PP_SEQ_E...
#include "boost/preprocessor/seq/limits/size_256.hpp" // for BOOST_PP_SEQ_S...
#include "caosdb/entity.h"                            // for Entity
#include "caosdb/logging.h"
#include "caosdb/entity/v1alpha1/main.grpc.pb.h" // for EntityTransactionSe...
#include "caosdb/entity/v1alpha1/main.pb.h"      // for Entity, RetrieveReq...
#include "caosdb/transaction_status.h"           // for TransactionStatus
#include "caosdb/status_code.h"                  // for StatusCode
#include "google/protobuf/util/json_util.h" // for MessageToJsonString, Jso...
#include <iterator>
// IWYU pragma: no_include <ext/alloc_traits.h>
#include <memory>    // for shared_ptr, unique_ptr
#include <stdexcept> // for out_of_range
#include <string>    // for string
#include <utility>   // for move
#include <vector>    // for vector

/**
 * Do all necessary checks and assure that another retrieval (by id or by
 * query) can be added as a sub-request to a transaction.
 */
#define ASSERT_CAN_ADD_RETRIEVAL                                               \
  if (!IsStatus(TransactionStatus::INITIAL()) &&                               \
      !IsStatus(TransactionStatus::GO_ON())) {                                 \
    return StatusCode::TRANSACTION_STATUS_ERROR;                               \
  }                                                                            \
  switch (this->transaction_type) {                                            \
  case NONE:                                                                   \
    this->transaction_type = TransactionType::READ_ONLY;                       \
    break;                                                                     \
  case READ_ONLY:                                                              \
    break;                                                                     \
  case MIXED_READ_AND_WRITE:                                                   \
    break;                                                                     \
  default:                                                                     \
    CAOSDB_LOG_ERROR_AND_RETURN_STATUS(                                        \
      logger_name, StatusCode::TRANSACTION_TYPE_ERROR,                         \
      "You cannot add a retrieval to this transaction because it has the "     \
      "wrong TransactionType.")                                                \
  }

/**
 * Do all necessary checks and assure that another retrieval (by id or by
 * query) can be added as a sub-request to a transaction.
 */
#define ASSERT_CAN_ADD_QUERY                                                   \
  ASSERT_CAN_ADD_RETRIEVAL                                                     \
  if (this->has_query) {                                                       \
    CAOSDB_LOG_ERROR_AND_RETURN_STATUS(                                        \
      logger_name, StatusCode::UNSUPPORTED_FEATURE,                            \
      "Currently the number of queries which can be processed in a single "    \
      "transaction is limitted to one.");                                      \
  }

/**
 * Do all necessary checks and assure that another deletion can be added as a
 * sub-request to a transaction.
 */
#define ASSERT_CAN_ADD_DELETION                                                \
  if (!IsStatus(TransactionStatus::INITIAL()) &&                               \
      !IsStatus(TransactionStatus::GO_ON())) {                                 \
    return StatusCode::TRANSACTION_STATUS_ERROR;                               \
  }                                                                            \
  switch (this->transaction_type) {                                            \
  case NONE:                                                                   \
    this->transaction_type = TransactionType::DELETE;                          \
  case DELETE:                                                                 \
  case MIXED_WRITE:                                                            \
  case MIXED_READ_AND_WRITE:                                                   \
    break;                                                                     \
  default:                                                                     \
    CAOSDB_LOG_ERROR_AND_RETURN_STATUS(                                        \
      logger_name, StatusCode::TRANSACTION_TYPE_ERROR,                         \
      "You cannot add a deletion to this transaction because it has the "      \
      "wrong TransactionType.")                                                \
  }

/**
 * Do all necessary checks and assure that another insertion can be added as a
 * sub-request to a transaction.
 */
#define ASSERT_CAN_ADD_INSERTION                                               \
  if (!IsStatus(TransactionStatus::INITIAL()) &&                               \
      !IsStatus(TransactionStatus::GO_ON())) {                                 \
    return StatusCode::TRANSACTION_STATUS_ERROR;                               \
  }                                                                            \
  switch (this->transaction_type) {                                            \
  case NONE:                                                                   \
    this->transaction_type = TransactionType::INSERT;                          \
  case INSERT:                                                                 \
  case MIXED_WRITE:                                                            \
  case MIXED_READ_AND_WRITE:                                                   \
    break;                                                                     \
  default:                                                                     \
    CAOSDB_LOG_ERROR_AND_RETURN_STATUS(                                        \
      logger_name, StatusCode::TRANSACTION_TYPE_ERROR,                         \
      "You cannot add an insertion to this transaction because it has the "    \
      "wrong TransactionType.")                                                \
  }

/**
 * Do all necessary checks and assure that another update can be added as a
 * sub-request to a transaction.
 */
#define ASSERT_CAN_ADD_UPDATE                                                  \
  if (!IsStatus(TransactionStatus::INITIAL()) &&                               \
      !IsStatus(TransactionStatus::GO_ON())) {                                 \
    return StatusCode::TRANSACTION_STATUS_ERROR;                               \
  }                                                                            \
  switch (this->transaction_type) {                                            \
  case NONE:                                                                   \
    this->transaction_type = TransactionType::INSERT;                          \
  case INSERT:                                                                 \
  case MIXED_WRITE:                                                            \
  case MIXED_READ_AND_WRITE:                                                   \
    break;                                                                     \
  default:                                                                     \
    CAOSDB_LOG_ERROR_AND_RETURN_STATUS(                                        \
      logger_name, StatusCode::TRANSACTION_TYPE_ERROR,                         \
      "You cannot add an update to this transaction because it has the "       \
      "wrong TransactionType.")                                                \
  }                                                                            \
  if (!entity->HasId()) {                                                      \
    CAOSDB_LOG_ERROR_AND_RETURN_STATUS(                                        \
      logger_name, StatusCode::ORIGINAL_ENTITY_MISSING_ID,                     \
      "You cannot update this entity without any id. Probably you did not "    \
      "retrieve it first? Entity updates should always start with the "        \
      "retrieval of the existing entity which may then be changed.")           \
  }

/**
 * @brief Creation and execution of transactions.
 * @author Timm Fitschen
 * @date 2021-08-05
 */
namespace caosdb::transaction {
using caosdb::entity::Entity;
using caosdb::entity::FileDescriptor;
using caosdb::entity::v1alpha1::EntityResponse;
using caosdb::entity::v1alpha1::EntityTransactionService;
using caosdb::entity::v1alpha1::FileDownloadRequest;
using caosdb::entity::v1alpha1::FileDownloadResponse;
using caosdb::entity::v1alpha1::FileTransmissionId;
using caosdb::entity::v1alpha1::FileTransmissionService;
using caosdb::entity::v1alpha1::FileUploadRequest;
using caosdb::entity::v1alpha1::FileUploadResponse;
using caosdb::entity::v1alpha1::IdResponse;
using caosdb::entity::v1alpha1::MultiTransactionRequest;
using caosdb::entity::v1alpha1::MultiTransactionResponse;
using caosdb::entity::v1alpha1::RegisterFileUploadRequest;
using caosdb::entity::v1alpha1::RegisterFileUploadResponse;
using caosdb::transaction::TransactionStatus;
using WrappedResponseCase =
  caosdb::entity::v1alpha1::TransactionResponse::WrappedResponseCase;

class Transaction;

static const std::string logger_name = "caosdb::transaction";

/**
 * Abstract base class for the results of a Transaction.
 */
template <class T> class ResultSet {
  class iterator;

public:
  virtual ~ResultSet() = default;
  [[nodiscard]] virtual auto Size() const noexcept -> int = 0;
  [[nodiscard]] virtual auto At(const int index) const -> const T & = 0;
  auto begin() const -> iterator;
  auto end() const -> iterator;

private:
  class iterator : public std::iterator<std::output_iterator_tag, T> {
  public:
    explicit iterator(const ResultSet<T> *result_set, int index = 0);
    auto operator*() const -> const T &;
    auto operator++() -> iterator &;
    auto operator++(int) -> iterator;
    auto operator!=(const iterator &rhs) const -> bool;

  private:
    int current_index = 0;
    const ResultSet<T> *result_set;
  };
};

template <class T>
ResultSet<T>::iterator::iterator(const ResultSet *result_set_param, int index)
  : current_index(index), result_set(result_set_param) {}

template <class T> auto ResultSet<T>::iterator::operator*() const -> const T & {
  return this->result_set->At(current_index);
}

template <class T>
auto ResultSet<T>::iterator::operator++() -> ResultSet<T>::iterator & {
  current_index++;
  return *this;
}

template <class T>
auto ResultSet<T>::iterator::operator++(int) -> ResultSet<T>::iterator {
  iterator tmp(*this);
  operator++();
  return tmp;
}

template <class T>
auto ResultSet<T>::iterator::operator!=(const iterator &rhs) const -> bool {
  return this->current_index != rhs.current_index;
}

template <class T> auto ResultSet<T>::begin() const -> ResultSet<T>::iterator {
  return ResultSet<T>::iterator(this, 0);
}

template <class T> auto ResultSet<T>::end() const -> ResultSet<T>::iterator {
  return ResultSet<T>::iterator(this, Size());
}

template <class T> class AbstractMultiResultSet : public ResultSet<T> {
public:
  virtual ~AbstractMultiResultSet() = default;
  inline explicit AbstractMultiResultSet(
    std::vector<std::unique_ptr<T>> result_set)
    : items(std::move(result_set)) {}
  [[nodiscard]] inline auto Size() const noexcept -> int override {
    return this->items.size();
  }
  [[nodiscard]] inline auto At(const int index) const -> const T & override {
    return *(this->items.at(index));
  }

protected:
  std::vector<std::unique_ptr<T>> items;
};

/**
 * Container of files which have been downloaded during a transaction.
 */
class FilesResultSet : public AbstractMultiResultSet<FileDescriptor> {
public:
  ~FilesResultSet() = default;
  explicit FilesResultSet(
    std::vector<std::unique_ptr<FileDescriptor>> result_set);
};

/**
 * Container with results of a transaction.
 *
 * In contrast to UniqueResult, this one can also hold multiple entities or zero
 * entities.
 */
class MultiResultSet : public AbstractMultiResultSet<Entity> {
public:
  ~MultiResultSet() = default;
  explicit MultiResultSet(std::vector<std::unique_ptr<Entity>> result_set);
};

/**
 * Container with the single result of a transaction.
 *
 * In contrast to MultiResultSet, this one guarantees to hold exactly one
 * entity.
 */
class UniqueResult : public ResultSet<Entity> {
public:
  ~UniqueResult() = default;
  explicit inline UniqueResult(EntityResponse *entityResponse)
    : entity(new Entity(entityResponse)){};
  explicit inline UniqueResult(IdResponse *idResponse)
    : entity(new Entity(idResponse)){};
  [[nodiscard]] auto GetEntity() const -> const Entity &;
  [[nodiscard]] inline auto Size() const noexcept -> int override { return 1; }
  [[nodiscard]] inline auto At(const int index) const
    -> const Entity & override {
    if (index != 0) {
      throw std::out_of_range("Index out of range. Length is 1.");
    }
    return *(this->entity);
  }

  friend Transaction;

private:
  std::unique_ptr<Entity> entity;
};

/**
 * @brief Create a transaction via `CaosDBConnection.createTransaction()`
 */
class Transaction {
public:
  /**
   * The transaction type restricts the kind of sub-transaction which may be
   * added to a transaction (insertion, update, deletion, retrieval).
   *
   * @note MIXED_READ_AND_WRITE and MIXED_WRITE transaction are not supported
   * yet.
   */
  enum TransactionType {
    NONE,                //!< Unspecified or not specified yet.
    READ_ONLY,           //!< Only retrievals (by id, by query)
    INSERT,              //!< Only insertions
    UPDATE,              //!< Only updates
    DELETE,              //!< Only deletions
    MIXED_WRITE,         //!< Only insertions, deletions, updates
    MIXED_READ_AND_WRITE //!< all kind of transaction.
  };
  Transaction(std::shared_ptr<EntityTransactionService::Stub> entity_service,
              std::shared_ptr<FileTransmissionService::Stub> file_service);

  /**
   * Add an entity id to this transaction for retrieval and also download the
   * file.
   *
   * If the entity doesn't have a file a warning is appended.
   *
   * If the file cannot be downloaded due to unsufficient permissions an error
   * is appended.
   */
  auto RetrieveAndDownloadFilesById(const std::string &id,
                                    const std::string &local_path) noexcept
    -> StatusCode;

  /**
   * Return a ResultSet<FileDescriptor>.
   *
   * The file result set is empty until the transaction terminates.
   */
  auto GetDownloadedFiles() const noexcept -> const FilesResultSet &;

  /**
   * Add an entity id to this transaction for retrieval.
   *
   * The retrieval is being processed when the Execute() or
   * ExecuteAsynchronously() methods of this transaction are called.
   */
  auto RetrieveById(const std::string &id) noexcept -> StatusCode;

  /**
   * Add all entity ids to this transaction for retrieval.
   */
  template <class InputIterator>
  inline auto RetrieveById(InputIterator begin, InputIterator end) noexcept
    -> StatusCode;

  /**
   * Add a query to this transaction.
   *
   * Currently, only FIND and COUNT queries are supported.
   */
  auto Query(const std::string &query) noexcept -> StatusCode;

  /**
   * Add the entity to this transaction for an insertion.
   *
   * The insertion is being processed when the Execute() or
   * ExecuteAsynchronously() methods of this transaction are called.
   *
   * Changing the entity afterwards results in undefined behavior.
   */
  auto InsertEntity(Entity *entity) noexcept -> StatusCode;

  /**
   * Add the entity to this transaction for an update.
   *
   * The update is being processed when the Execute() or
   * ExecuteAsynchronously() methods of this transaction are called.
   *
   * Changing the entity afterwards results in undefined behavior.
   */
  auto UpdateEntity(Entity *entity) noexcept -> StatusCode;

  /**
   * Add an entity id to this transaction for deletion.
   *
   * The deletion is being processed when the Execute() or
   * ExecuteAsynchronously() methods of this transaction are called.
   */
  auto DeleteById(const std::string &id) noexcept -> StatusCode;

  inline auto IsStatus(const TransactionStatus &status) const noexcept -> bool {
    return this->status.GetCode() == status.GetCode();
  };

  /**
   * Execute this transaction in blocking mode and return the status.
   */
  auto Execute() -> TransactionStatus;

  /**
   * Execute this transaction in non-blocking mode and return immediately.
   *
   * A client may request the current status at any time via GetStatus().
   *
   * Use WaitForIt() to join the back-ground execution of this transaction.
   */
  auto ExecuteAsynchronously() noexcept -> StatusCode;

  /**
   * Join the background execution and return the status when the execution
   * terminates.
   *
   * Use this after ExecuteAsynchronously().
   */
  [[nodiscard]] auto WaitForIt() const noexcept -> TransactionStatus;

  /**
   * Return the current status of the transaction.
   */
  [[nodiscard]] inline auto GetStatus() const noexcept -> TransactionStatus {
    return this->status;
  }

  [[nodiscard]] inline auto GetResultSet() const noexcept
    -> const ResultSet<Entity> & {
    const ResultSet<Entity> *result_set = this->result_set.get();
    return *result_set;
  }

  /**
   * Return the result of a count query
   *
   * Only meaningful if there was exactly one COUNT query executed in
   * this transaction. In all other cases, the return value will be
   * -1.
   */
  [[nodiscard]] inline auto GetCountResult() const noexcept -> long {
    return query_count;
  }

  /**
   * Return the number of sub-requests in this transaction.
   *
   * This is meant for debugging because the number of sub-requests is a
   * GRPC-API detail.
   */
  [[nodiscard]] inline auto GetRequestCount() const -> int {
    return this->request->requests_size();
  }

  /**
   * Get a JSON representation of the response.
   *
   * For debugging.
   */
  inline auto ResponseToString() const -> const std::string {
    google::protobuf::util::JsonOptions options;
    std::string out;
    google::protobuf::util::MessageToJsonString(*this->response, &out, options);
    return out;
  }

  /**
   * Get a JSON representation of the request.
   *
   * For debugging.
   */
  inline auto RequestToString() const -> const std::string {
    google::protobuf::util::JsonOptions options;
    std::string out;
    google::protobuf::util::MessageToJsonString(*this->request, &out, options);
    return out;
  }

  std::vector<FileDescriptor> upload_files;
  std::map<std::string, FileDescriptor> download_files;

private:
  auto RegisterUploadFile(RegisterFileUploadResponse *response) -> void;
  auto UploadFile(FileUploadResponse *response,
                  const FileDescriptor &file_descriptor,
                  const std::string &registration_id) -> void;
  auto DownloadFile(FileDownloadResponse *response,
                    const FileTransmissionId &file_transmission_id) -> void;
  bool has_query = false;
  TransactionType transaction_type = TransactionType::NONE;
  mutable std::unique_ptr<ResultSet<Entity>> result_set;
  mutable TransactionStatus status = TransactionStatus::INITIAL();
  std::shared_ptr<EntityTransactionService::Stub> entity_service;
  std::shared_ptr<FileTransmissionService::Stub> file_service;
  MultiTransactionRequest *request;
  mutable MultiTransactionResponse *response;
  std::string error_message;
  mutable long query_count;
};

template <class InputIterator>
inline auto Transaction::RetrieveById(InputIterator begin,
                                      InputIterator end) noexcept
  -> StatusCode {
  ASSERT_CAN_ADD_RETRIEVAL

  auto next = begin;
  while (next != end) {
    auto *sub_request = this->request->add_requests();
    sub_request->mutable_retrieve_request()->set_id(*next);
    next = std::next(next);
  }

  this->status = TransactionStatus::GO_ON();
  return this->status.GetCode();
}

} // namespace caosdb::transaction
#endif
