#include "caosdb/filestreaming/DownloadRequestHandler.h"
#include "caosdb/protobuf_helper.h"
#include "caosdb/exceptions.h"
#include "caosdb/status_code.h"
#include "caosdb/logging.h"

#include <iostream>

namespace FileExchange {
using caosdb::StatusCode;
using caosdb::exceptions::AuthenticationError;
using caosdb::exceptions::ConnectionError;
using caosdb::exceptions::Exception;
using caosdb::utility::get_arena;
using google::protobuf::Arena;

DownloadRequestHandler::DownloadRequestHandler(
  HandlerTag tag, FileTransmissionService::Stub *stub,
  grpc::CompletionQueue *cq, const FileDescriptor &file_descriptor)
  : tag_(tag), stub_(stub), cq_(cq), state_(CallState::NewCall),
    file_descriptor_(file_descriptor), bytesReceived_(0),
    request_(Arena::CreateMessage<FileDownloadRequest>(get_arena())),
    response_(Arena::CreateMessage<FileDownloadResponse>(get_arena())) {
  this->onNext(true);
}

bool DownloadRequestHandler::onNext(bool ok) {
  try {
    if (ok) {
      if (state_ == CallState::NewCall) {
        this->handleNewCallState();
      } else if (state_ == CallState::SendingRequest) {
        this->handleSendingRequestState();
      } else if (state_ == CallState::ReceivingFile) {
        this->handleReceivingFileState();
      } else if (state_ == CallState::CallComplete) {
        this->handleCallCompleteState();
        return false; // TODO comment
      }
    } else {
      state_ = CallState::CallComplete;
      rpc_->Finish(&status_, tag_);
    }

    return true;
  } catch (Exception &e) {
    throw;
  } catch (std::exception &e) {
    CAOSDB_LOG_ERROR(logger_name) << "Download processing error: " << e.what();
    throw;
  } catch (...) {
    CAOSDB_LOG_ERROR(logger_name)
      << "Download processing error: unknown exception caught";
    throw;
  }

  if (state_ == CallState::NewCall) {
    // TODO comment
    return false;
  }

  ctx_.TryCancel();

  return true;
}

void DownloadRequestHandler::cancel() { ctx_.TryCancel(); }

void DownloadRequestHandler::handleNewCallState() {
  CAOSDB_LOG_TRACE(logger_name)
    << "Enter DownloadRequestHandler::handleNewCallState. local_path = "
    << file_descriptor_.local_path
    << ", download_id = " << file_descriptor_.file_transmission_id;
  const std::size_t ServerDefaultChunkSize = 0;
  fileWriter_ = std::make_unique<FileWriter>(file_descriptor_.local_path);

  CAOSDB_DEBUG_MESSAGE_STRING(*request_, request_out)
  CAOSDB_LOG_TRACE(logger_name) << "HERE1" << request_out;
  CAOSDB_DEBUG_MESSAGE_STRING(*(file_descriptor_.file_transmission_id),
                              file_transmission_id_out)
  CAOSDB_LOG_TRACE(logger_name) << "HERE2" << file_transmission_id_out;
  request_->mutable_file_transmission_id()->CopyFrom(
    *(file_descriptor_.file_transmission_id));
  CAOSDB_LOG_TRACE(logger_name) << "HERE2";

  rpc_ = stub_->PrepareAsyncFileDownload(&ctx_, *request_, cq_);
  CAOSDB_LOG_TRACE(logger_name) << "HERE3";

  state_ = CallState::SendingRequest;
  rpc_->StartCall(tag_);
  CAOSDB_LOG_TRACE(logger_name)
    << "Leave DownloadRequestHandler::handleNewCallState";
}

void DownloadRequestHandler::handleSendingRequestState() {
  CAOSDB_LOG_TRACE(logger_name)
    << "Enter DownloadRequestHandler::handleSendingRequestState";
  state_ = CallState::ReceivingFile;
  rpc_->Read(response_, tag_);
  CAOSDB_LOG_TRACE(logger_name)
    << "Leave DownloadRequestHandler::handleSendingRequestState";
}

void DownloadRequestHandler::handleReceivingFileState() {
  CAOSDB_LOG_TRACE(logger_name)
    << "Enter DownloadRequestHandler::handleReceivingFileState";
  if (response_->has_chunk()) {
    auto &chunkData = response_->chunk().data();
    if (chunkData.empty()) {
      CAOSDB_LOG_DEBUG(logger_name) << "Received an empty FileChunk, ignoring";
    } else {
      fileWriter_->write(chunkData);
      bytesReceived_ += chunkData.size();
    }

    state_ = CallState::ReceivingFile;
    response_->Clear();
    rpc_->Read(response_, tag_);
  } else {
    throw std::runtime_error("File chunk expected");
  }
  CAOSDB_LOG_TRACE(logger_name)
    << "Leave DownloadRequestHandler::handleReceivingFileState";
}

void DownloadRequestHandler::handleCallCompleteState() {
  CAOSDB_LOG_TRACE(logger_name)
    << "Enter DownloadRequestHandler::handleCallCompleteState";
  switch (status_.error_code()) {
  case grpc::OK:
    CAOSDB_LOG_INFO(logger_name) << "[" << file_descriptor_.local_path
                                 << "]: download complete: " << bytesReceived_
                                 << " bytes received" << std::endl;
    break;

  case grpc::UNAUTHENTICATED:
    throw AuthenticationError(status_.error_message());
  case grpc::UNAVAILABLE:
    throw ConnectionError(status_.error_message());
  default:
    throw Exception(StatusCode::GENERIC_RPC_ERROR,
                    "GRPC error code " + std::to_string(status_.error_code()) +
                      " - " + status_.error_message());
  }
  CAOSDB_LOG_TRACE(logger_name)
    << "Leave DownloadRequestHandler::handleCallCompleteState";
}

} // namespace FileExchange
