/*
 * This file is a part of the CaosDB Project.
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */
#include "caosdb/transaction.h"
#include "caosdb/entity/v1alpha1/main.grpc.pb.h" // for EntityTransactionS...
#include "caosdb/entity/v1alpha1/main.pb.h"      // for SingleRetrieveRequest
#include "caosdb/filestreaming/Client.h"
#include "caosdb/logging.h"
#include "caosdb/protobuf_helper.h"               // for get_arena
#include "caosdb/status_code.h"                   // for StatusCode, AUTHEN...
#include "google/protobuf/arena.h"                // for Arena
#include "grpcpp/grpcpp.h"                        // for CompletionQueue
#include "grpcpp/impl/codegen/async_unary_call.h" // for ClientAsyncRespons...
#include "grpcpp/impl/codegen/client_context.h"   // for ClientContext
#include "grpcpp/impl/codegen/completion_queue.h" // for CompletionQueue
#include "grpcpp/impl/codegen/status.h"           // for Status
#include "grpcpp/impl/codegen/status_code_enum.h" // for StatusCode, UNAUTH...
#include <cassert>                                // for assert
#include <google/protobuf/util/json_util.h> // for MessageToJsonString, Jso...
#include <map>                              // for map
#include <memory>                           // for allocator, unique_ptr
#include <stdexcept>                        // for out_of_range
#include <utility>                          // for move

namespace caosdb {

// TODO(tf) move to another source file.
auto get_status_description(int code) -> const std::string & {
  static const std::string MISSING_DESCRIPTION = "MISSING DESCRIPTION";
  static const std::map<int, std::string> descriptions = {
    {StatusCode::INITIAL,
     "The transaction has just been intialized. It has not been executed yet "
     "and clients can still change it and add sub-transactions."},
    {StatusCode::GO_ON,
     "The transaction has a transaction_type yet and clients may add matching "
     "sub-transaction or execute it right-away."},
    {StatusCode::READY,
     "The transaction is ready for execution and cannot be changed anymore."},
    {StatusCode::EXECUTING, "The transaction is currently being executed."},
    {StatusCode::SUCCESS, "The action was successful"},
    {StatusCode::CONNECTION_ERROR,
     "The attempt to execute this transaction was not successful because the "
     "connection to the server could not be established."},
    {StatusCode::AUTHENTICATION_ERROR,
     "The attempt to execute this transaction has not been executed at all "
     "because the authentication did not succeed."},
    {StatusCode::GENERIC_RPC_ERROR,
     "The attempt to execute this transaction was not successful because an "
     "error occured in the transport or RPC protocol layer."},
    {StatusCode::GENERIC_ERROR,
     "An error occured. Please review the logs for more information."},
    {StatusCode::GENERIC_TRANSACTION_ERROR,
     "The transaction terminated unsuccessfully with transaction errors."},
    {StatusCode::CONFIGURATION_ERROR,
     "An error occurred during the configuration of the ConfigurationManager."},
    {StatusCode::UNKNOWN_CONNECTION_ERROR,
     "The ConnectionManager does not know any connection of this name."},
    {StatusCode::TRANSACTION_STATUS_ERROR,
     "The Transaction is in a wrong state for the attempted action."},
    {StatusCode::TRANSACTION_TYPE_ERROR,
     "The Transaction has a transaction type which does not allow the "
     "attempted action."},
    {StatusCode::ORIGINAL_ENTITY_MISSING_ID,
     "The attempt to update this entity failed because this entity does not "
     "have "
     "an id. This is the case when you did not retrieve it before applying any "
     "changes and instantiated the Entity class explicitely."},
    {StatusCode::NOT_A_FILE_ENTITY, "You can only add files to file entities."},
    {StatusCode::PATH_IS_A_DIRECTORY, "The given path is a directory."},
    {StatusCode::FILE_DOES_NOT_EXIST_LOCALLY,
     "The file does not not exist in the local file system."},
    {StatusCode::FILE_DOWNLOAD_ERROR,
     "The transaction failed during the download of the files"},
    {StatusCode::FILE_UPLOAD_ERROR,
     "The transaction failed during the upload of the files"},
    {StatusCode::UNSUPPORTED_FEATURE,
     "This feature is not available in the this client implementation."}};
  try {
    return descriptions.at(code);
  } catch (const std::out_of_range &exc) {
    return MISSING_DESCRIPTION;
  }
}

} // namespace caosdb

namespace caosdb::transaction {
using caosdb::entity::v1alpha1::EntityTransactionService;
using caosdb::entity::v1alpha1::FileTransmissionService;
using caosdb::entity::v1alpha1::MultiTransactionRequest;
using caosdb::entity::v1alpha1::MultiTransactionResponse;
using TransactionResponseCase =
  caosdb::entity::v1alpha1::TransactionResponse::WrappedResponseCase;
using QueryResponseCase =
  caosdb::entity::v1alpha1::RetrieveResponse::WrappedResponseCase;
using caosdb::utility::get_arena;
using grpc::ClientAsyncResponseReader;
using ProtoEntity = caosdb::entity::v1alpha1::Entity;
using FileExchange::FileExchangeClient;
using google::protobuf::Arena;
using grpc::CompletionQueue;

MultiResultSet::MultiResultSet(std::vector<std::unique_ptr<Entity>> result_set)
  : AbstractMultiResultSet<Entity>(std::move(result_set)) {}

[[nodiscard]] auto UniqueResult::GetEntity() const -> const Entity & {
  const Entity *result = this->entity.get();
  return *result;
}

Transaction::Transaction(
  std::shared_ptr<EntityTransactionService::Stub> entity_service,
  std::shared_ptr<FileTransmissionService::Stub> file_service)
  : request(Arena::CreateMessage<MultiTransactionRequest>(get_arena())),
    response(Arena::CreateMessage<MultiTransactionResponse>(get_arena())) {
  this->entity_service = std::move(entity_service);
  this->file_service = std::move(file_service);
  this->query_count = -1;
}

auto Transaction::RetrieveById(const std::string &id) noexcept -> StatusCode {
  ASSERT_CAN_ADD_RETRIEVAL

  auto *sub_request = this->request->add_requests();
  sub_request->mutable_retrieve_request()->set_id(id);

  this->status = TransactionStatus::GO_ON();
  return this->status.GetCode();
}

auto Transaction::RetrieveAndDownloadFilesById(
  const std::string &id, const std::string &local_path) noexcept -> StatusCode {
  ASSERT_CAN_ADD_RETRIEVAL

  auto *retrieve_request =
    this->request->add_requests()->mutable_retrieve_request();
  retrieve_request->set_id(id);
  retrieve_request->set_register_file_download(true);

  download_files[id].local_path = local_path;

  this->status = TransactionStatus::GO_ON();
  return this->status.GetCode();
}

auto Transaction::Query(const std::string &query) noexcept -> StatusCode {
  ASSERT_CAN_ADD_QUERY

  // currently, the server can only process one query at a time (but mixed with
  // other retrievals).
  this->has_query = true;
  auto *sub_request = this->request->add_requests();
  sub_request->mutable_retrieve_request()->mutable_query()->set_query(query);

  this->status = TransactionStatus::GO_ON();
  return this->status.GetCode();
}

auto Transaction::DeleteById(const std::string &id) noexcept -> StatusCode {
  ASSERT_CAN_ADD_DELETION

  auto *sub_request = this->request->add_requests();
  sub_request->mutable_delete_request()->set_id(id);

  this->status = TransactionStatus::GO_ON();
  return this->status.GetCode();
}

auto Transaction::InsertEntity(Entity *entity) noexcept -> StatusCode {
  ASSERT_CAN_ADD_INSERTION

  auto *entity_request = this->request->add_requests()
                           ->mutable_insert_request()
                           ->mutable_entity_request();
  auto *proto_entity = entity_request->mutable_entity();

  // copy the original entity for the transaction
  entity->CopyTo(proto_entity);
  if (entity->HasFile()) {
    auto *file_transmission_id = entity_request->mutable_upload_id();
    entity->SetFileTransmissionId(file_transmission_id);
    upload_files.push_back(entity->GetFileDescriptor());
  }
  this->status = TransactionStatus::READY();
  return this->status.GetCode();
}

auto Transaction::UpdateEntity(Entity *entity) noexcept -> StatusCode {
  ASSERT_CAN_ADD_UPDATE

  auto *entity_request = this->request->add_requests()
                           ->mutable_update_request()
                           ->mutable_entity_request();
  auto *proto_entity = entity_request->mutable_entity();

  entity->CopyTo(proto_entity);
  if (entity->HasFile()) {
    auto *file_transmission_id = entity_request->mutable_upload_id();
    entity->SetFileTransmissionId(file_transmission_id);
    upload_files.push_back(entity->GetFileDescriptor());
  }
  this->status = TransactionStatus::READY();
  return this->status.GetCode();
}

auto Transaction::Execute() -> TransactionStatus {
  auto status_code = ExecuteAsynchronously();
  TransactionStatus::ThrowExceptionIfError(
    status_code, caosdb::get_status_description(status_code));
  auto status = WaitForIt();
  status.ThrowExceptionIfError();
  return status;
}

auto Transaction::ExecuteAsynchronously() noexcept -> StatusCode {
  if (!IsStatus(TransactionStatus::READY()) &&
      !IsStatus(TransactionStatus::GO_ON())) {
    return StatusCode::TRANSACTION_STATUS_ERROR;
  }
  switch (this->transaction_type) {
  case MIXED_WRITE:
    CAOSDB_LOG_ERROR_AND_RETURN_STATUS(
      logger_name, StatusCode::UNSUPPORTED_FEATURE,
      "MIXED_WRITE UNSUPPORTED: The current implementation does not support "
      "mixed write transactions (containing insertions, deletions, and updates "
      "in one transaction).")
  case MIXED_READ_AND_WRITE:
    CAOSDB_LOG_ERROR_AND_RETURN_STATUS(
      logger_name, StatusCode::UNSUPPORTED_FEATURE,
      "MIXED_WRITE UNSUPPORTED: The current implementation does not support "
      "mixed read and write transactions (containing retrievals, insertions, "
      "deletions, and updates in one transaction).")
  default:
    break;
  }
  this->status = TransactionStatus::EXECUTING();

  // upload files first
  if (upload_files.size() > 0) {
    auto *registration_response =
      Arena::CreateMessage<RegisterFileUploadResponse>(get_arena());
    RegisterUploadFile(registration_response);

    // TODO if(registration_response.status != REGISTRATION_STATUS_ACCEPTED){
    //   return StatusCode::FILE_UPLOAD_FAILED
    // }
    FileExchangeClient upload_client(file_service);
    for (auto file_descriptor : upload_files) {
      file_descriptor.file_transmission_id->set_registration_id(
        registration_response->registration_id());
      CAOSDB_LOG_INFO(logger_name)
        << "Uploading " << file_descriptor.local_path;
      auto file_upload_status = upload_client.upload(file_descriptor);
      if (!file_upload_status == StatusCode::SUCCESS) {
        this->status = TransactionStatus::FILE_UPLOAD_ERROR();
        return StatusCode::EXECUTING;
      }
    }
  }

  grpc::Status grpc_status;
  CompletionQueue cq;

  grpc::ClientContext context;
  std::unique_ptr<ClientAsyncResponseReader<MultiTransactionResponse>> rpc(
    this->entity_service->PrepareAsyncMultiTransaction(&context,
                                                       *(this->request), &cq));
  rpc->StartCall();

  int tag = 1;
  void *send_tag = static_cast<void *>(&tag);
  rpc->Finish(this->response, &grpc_status, send_tag);
  void *recv_tag = nullptr;
  bool ok = false;

  // TODO(tf) make this actually asynchronous by moving this to WaitForIt()
  cq.Next(&recv_tag, &ok);
  assert(recv_tag == send_tag);
  assert(ok);

  if (!grpc_status.ok()) {
    switch (grpc_status.error_code()) {
    case grpc::StatusCode::UNAUTHENTICATED:
      this->status = TransactionStatus::AUTHENTICATION_ERROR();
      break;
    case grpc::StatusCode::UNAVAILABLE:
      this->status = TransactionStatus::CONNECTION_ERROR();
      break;
    default:
      auto error_details = std::to_string(grpc_status.error_code()) + " - " +
                           grpc_status.error_message();
      this->status = TransactionStatus::RPC_ERROR(error_details);
    }
  } else {
    this->status = TransactionStatus::SUCCESS();
  }

  // file download afterwards
  if (status.GetCode() == StatusCode::SUCCESS && download_files.size() > 0) {
    // run over all retrieved entities and get the download_id
    for (auto sub_response : *(response->mutable_responses())) {
      if (sub_response.wrapped_response_case() ==
          TransactionResponseCase::kRetrieveResponse) {
        if (sub_response.retrieve_response()
              .entity_response()
              .has_download_id()) {
          auto entity_response =
            sub_response.mutable_retrieve_response()->mutable_entity_response();
          auto entity_id = entity_response->entity().id();
          download_files[entity_id].file_transmission_id =
            entity_response->release_download_id();
          // TODO(tf) handle error
        }
      }
    }

    auto *registration_response =
      Arena::CreateMessage<RegisterFileUploadResponse>(get_arena());
    RegisterUploadFile(registration_response);

    // TODO if(registration_response.status != REGISTRATION_STATUS_ACCEPTED){
    //   return StatusCode::FILE_UPLOAD_FAILED
    // }
    FileExchangeClient download_client(file_service);
    for (auto item : download_files) {
      auto file_descriptor(item.second);
      CAOSDB_DEBUG_MESSAGE_STRING(*file_descriptor.file_transmission_id, out)
      CAOSDB_LOG_INFO(logger_name)
        << "Downloading " << file_descriptor.local_path << ", " << out;
      auto file_download_status = download_client.download(file_descriptor);
      if (!file_download_status == StatusCode::SUCCESS) {
        this->status = TransactionStatus::FILE_DOWNLOAD_ERROR();
        // TODO (tf) handle multiple errors?
        return StatusCode::EXECUTING;
      }
    }
  }
  return StatusCode::EXECUTING;
}

auto Transaction::WaitForIt() const noexcept -> TransactionStatus {
  if (this->response->responses_size() == 1) {
    auto *responses = this->response->mutable_responses(0);
    switch (responses->wrapped_response_case()) {
    case WrappedResponseCase::kRetrieveResponse: {
      auto *retrieve_response = responses->mutable_retrieve_response();
      switch (retrieve_response->wrapped_response_case()) {
      case QueryResponseCase::kEntityResponse: {
        auto *entity = retrieve_response->release_entity_response();
        auto id = entity->entity().id();
        if (!entity->errors().empty()) {
          this->status = TransactionStatus::TRANSACTION_ERROR(
            "The request returned with errors.");
        }

        this->result_set = std::make_unique<UniqueResult>(entity);
        if (!id.empty() && download_files.count(id) == 1) {
          const auto &local_path = download_files.at(id).local_path;
          const auto &unique_result_set =
            dynamic_cast<const UniqueResult &>(*result_set.get());
          unique_result_set.entity->SetLocalPath(local_path);
        }
      } break;
      case QueryResponseCase::kSelectResult: {
        CAOSDB_LOG_ERROR(logger_name) << "Results of a SELECT query cannot be "
                                         "processed by this client yet.";
        // TODO(tf) Select queries
      } break;
      case QueryResponseCase::kCountResult: {
        this->query_count = retrieve_response->count_result();
        std::vector<std::unique_ptr<Entity>> entities;
        this->result_set =
          std::make_unique<MultiResultSet>(std::move(entities));
      } break;
      default:
        CAOSDB_LOG_FATAL(logger_name) << "Received invalid QueryResponseCase.";
        break;
      }
    } break;
    case WrappedResponseCase::kUpdateResponse: {
      auto *updatedIdResponse = responses->mutable_update_response();
      if (!updatedIdResponse->id_response().errors().empty()) {
        this->status = TransactionStatus::TRANSACTION_ERROR(
          "The request returned with errors.");
      }
      this->result_set = std::make_unique<UniqueResult>(
        updatedIdResponse->release_id_response());
    } break;
    case WrappedResponseCase::kInsertResponse: {
      auto *insertedIdResponse = responses->mutable_insert_response();
      if (!insertedIdResponse->id_response().errors().empty()) {
        this->status = TransactionStatus::TRANSACTION_ERROR(
          "The request returned with errors.");
      }
      this->result_set = std::make_unique<UniqueResult>(
        insertedIdResponse->release_id_response());
    } break;
    case WrappedResponseCase::kDeleteResponse: {
      auto *deletedIdResponse = responses->mutable_delete_response();
      if (!deletedIdResponse->id_response().errors().empty()) {
        this->status = TransactionStatus::TRANSACTION_ERROR(
          "The request returned with errors.");
      }
      this->result_set = std::make_unique<UniqueResult>(
        deletedIdResponse->release_id_response());
    } break;
    default:
      CAOSDB_LOG_FATAL(logger_name) << "Received invalid WrappedResponseCase.";
      break;
    }
  } else {
    auto *responses = this->response->mutable_responses();
    std::vector<std::unique_ptr<Entity>> entities;
    CAOSDB_LOG_DEBUG(logger_name)
      << "Number of subresponses: " << responses->size();
    for (auto sub_response : *responses) {
      switch (sub_response.wrapped_response_case()) {
      case TransactionResponseCase::kRetrieveResponse:
        entities.push_back(std::make_unique<Entity>(
          sub_response.mutable_retrieve_response()->release_entity_response()));
        break;
      case TransactionResponseCase::kInsertResponse:
        entities.push_back(std::make_unique<Entity>(
          sub_response.mutable_insert_response()->release_id_response()));
        break;
      case TransactionResponseCase::kDeleteResponse:
        entities.push_back(std::make_unique<Entity>(
          sub_response.mutable_delete_response()->release_id_response()));
        break;
      case TransactionResponseCase::kUpdateResponse:
        entities.push_back(std::make_unique<Entity>(
          sub_response.mutable_update_response()->release_id_response()));
        break;
      default:
        CAOSDB_LOG_FATAL(logger_name)
          << "Received invalid TransactionResponseCase.";
        break;
      }
    }
    for (auto &entity : entities) {
      auto id = entity->GetId();
      if (!id.empty() && download_files.count(id) == 1) {
        const auto &local_path = download_files.at(id).local_path;
        entity->SetLocalPath(local_path);
      }
    }
    this->result_set = std::make_unique<MultiResultSet>(std::move(entities));
  }

  return this->status;
}

auto Transaction::RegisterUploadFile(RegisterFileUploadResponse *response)
  -> void {
  grpc::Status grpc_status;
  CompletionQueue cq;

  RegisterFileUploadRequest request;

  grpc::ClientContext context;
  std::unique_ptr<ClientAsyncResponseReader<RegisterFileUploadResponse>> rpc(
    this->file_service->PrepareAsyncRegisterFileUpload(&context, request, &cq));

  rpc->StartCall();

  int tag = 1;
  void *send_tag = static_cast<void *>(&tag);
  rpc->Finish(response, &grpc_status, send_tag);
  void *recv_tag = nullptr;
  bool ok = false;

  cq.Next(&recv_tag, &ok);
  assert(recv_tag == send_tag);
  assert(ok);
}

// auto Transaction::UploadFile(FileUploadResponse *response,
// const FileDescriptor &file_descriptor,
// const std::string &registration_id) -> void {
// grpc::Status grpc_status;
// CompletionQueue cq;
// grpc::ClientContext context;

// FileUploadRequest request;
// auto *chunk = request.mutable_chunk();
// auto *file_transmission_id = chunk->mutable_file_transmission_id();
// file_transmission_id->set_registration_id(registration_id);
// file_transmission_id->set_file_id("test.txt");
// chunk->set_data("this is some data");

// std::unique_ptr<ClientAsyncResponseReader<FileUploadResponse>> rpc(
// this->file_service->PrepareAsyncFileUpload(&context, request, &cq));

// rpc->StartCall();

// int tag = 1;
// void *send_tag = static_cast<void *>(&tag);
// rpc->Finish(response, &grpc_status, send_tag);
// void *recv_tag = nullptr;
// bool ok = false;

// cq.Next(&recv_tag, &ok);
// assert(recv_tag == send_tag);
// assert(ok);
//}

// auto Transaction::DownloadFile(FileDownloadResponse *response,
// const FileTransmissionId &file_transmission_id)
//-> void {

// FileDownloadRequest request;
// request.mutable_file_transmission_id()->CopyFrom(file_transmission_id);
// grpc::Status grpc_status;
// CompletionQueue cq;

// grpc::ClientContext context;
// std::unique_ptr<ClientAsyncResponseReader<FileDownloadResponse>> rpc(
// this->file_service->PrepareAsyncFileDownload(&context, request, &cq));

// rpc->StartCall();

// int tag = 1;
// void *send_tag = static_cast<void *>(&tag);
// rpc->Finish(response, &grpc_status, send_tag);
// void *recv_tag = nullptr;
// bool ok = false;

// cq.Next(&recv_tag, &ok);
//}

} // namespace caosdb::transaction
